/****************************************************************
 *                  Clown IDS v1.0 by mimayin                   *
 *                                                              *
 *  simple IDS system designed for *bsd, but could probably be  *
 *              trivially modifed to run on linux               *
 *                                                              *
 ****************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <unistd.h>
#include <sys/time.h>
#include <string.h>

/******* CONFIGURATION OPTIONS ********/

#define IDS_LOG		"/var/log/ids.log"			/* ids log file */
#define IDS_CONF	"/etc/ids.conf"				/* directory config file */
#define KEY_FILE	"/.idskeyfile"				/* key file to allow editing */
#define VERBOSE		1					/* set to 1 for more startup info */
#define INTERVAL	3					/* check interval in minutes */
#define DB_PREFIX	"/.ids"					/* path to put md5 snapshot data !! MUST BE OUTSIDE PATHS IN ids.conf!! */
#define EMER_CMD	"/bin/emergency"			/* command or script to execute on breakin detection */
#define ADMIN_MAIL	"root@localhost"			/* email addr to mail changes to in case of breakin */
#define SUBJECT		"ClownIDS: !!Intrusion Detected!!"	/* subject for email to admin */

/**************************************/

FILE *logfile;
char data[128][128];
char snap[128][128];
char logbuf[1024];
int n=0;

void log(char *buff);

void banner()
{
	printf("\n");
	printf("*****************************************************\n");
	printf("*                                                   *\n");
	printf("*       ClownIDS v1.0 by mimayin@lsa.mine.nu        *\n");
	printf("*                                                   *\n");
	printf("*                http://lsa.mine.nu                 *\n");
	printf("*                                                   *\n");
	printf("*****************************************************\n\n");
}

void sig_exit()
{
}

void init_signals()
{
        signal(SIGHUP, sig_exit);
        signal(SIGINT, sig_exit);
        signal(SIGQUIT, sig_exit);
        signal(SIGILL, sig_exit);
        signal(SIGTRAP, sig_exit);
        signal(SIGIOT, sig_exit);
        signal(SIGBUS, sig_exit);
        signal(SIGFPE, sig_exit);
        signal(SIGKILL, sig_exit);
        signal(SIGUSR1, sig_exit);
        signal(SIGSEGV, sig_exit);
        signal(SIGUSR2, sig_exit);
        signal(SIGPIPE, sig_exit);
        signal(SIGALRM, sig_exit);
        signal(SIGTERM, sig_exit);
        signal(SIGCHLD, sig_exit);
        signal(SIGCONT, sig_exit);
        signal(SIGSTOP, sig_exit);
        signal(SIGTSTP, sig_exit);
        signal(SIGTTIN, sig_exit);
        signal(SIGTTOU, sig_exit);
        signal(SIGURG, sig_exit);
        signal(SIGXCPU, sig_exit);
        signal(SIGXFSZ, sig_exit);
        signal(SIGVTALRM, sig_exit);
        signal(SIGPROF, sig_exit);
        signal(SIGWINCH, sig_exit);
        signal(SIGIO, sig_exit);
}

int readconfig()
{
	FILE *fp;
	int i,x=0;
	char buffer[128];

	init_signals();

	for(i=0;i<128;i++) memset(data[i],0,128);
	memset(buffer,0,128);

	printf("   Init: Initialising ClownIDS ...\n");
	sprintf(logbuf,"Initialising ...");log(logbuf);

	if(chdir(DB_PREFIX)!=0)
	{
		printf("  Error: Snapshot directory [%s] does not exist\n",DB_PREFIX);
		sprintf(logbuf,"Snapshot directory [%s] does not exist [halted]",DB_PREFIX);
		log(logbuf);
		return 0;
	}

	if((fp = fopen(IDS_CONF,"r")) == NULL)
	{
		printf("  Error: Couldn't open %s\n",IDS_CONF);
		sprintf(logbuf,"Couldn't open config file %s [halted]",IDS_CONF);
		log(logbuf);
		return 0;
	}

	while((i=getc(fp))!=EOF)
	{
		if(i==10)
		{
			if(buffer[0]=='#')
			{
				memset(buffer,0,128);
				x=0;
			}
			else if(strlen(buffer)<2)
			{
				memset(buffer,0,128); x=0;
			}
			else if(chdir(buffer)!=0)
			{
				fclose(fp);
				printf("  Error: %s line %d: directory not found.\n",IDS_CONF,n+1);
				sprintf(logbuf,"%s line %d: directory not found [halted]",IDS_CONF,n+1);
				log(logbuf);
				return 0;
			}
			else if(strlen(buffer)>127)
			{
				printf("WARNING: %s line %d ignored: name exceeds 127 chars. [%s]\n",IDS_CONF,n+1,buffer);
				sprintf(logbuf, "WARNING: %s line %d ignored: name exceeds 127 chars [%s]\n",IDS_CONF,n+1,buffer);
				log(logbuf);
				memset(buffer,0,128); x=0;
			}
			else
			{
				strncpy(data[n++],buffer,strlen(buffer));
				sprintf(logbuf,"Adding [%s] to monitor list",buffer);log(logbuf);
				memset(buffer,0,128); x=0;
			}
		}
		else
		{
			buffer[x++] = i;
		}
	}

	fclose(fp);
	return 1;
}

void showdirs()
{
	int i;

	printf("   Init: Loaded: ");
	for(i=0;i<n;i++) printf("%s ",data[i]);
	printf("\n");
}

void log(char *buff)
{
        char buffer[2048];
        char day[3],month[3],year[5],hr[3],min[3],sec[3];
        time_t var1=time(NULL);
        struct tm *var2;

        memset(buffer,0,2048);
        memset(day,0,3);
        memset(month,0,3);
        memset(year,0,5);
        memset(hr,0,3);
        memset(min,0,3);
        memset(sec,0,3);

        var2=localtime(&var1);

        if(var2->tm_mday<10)    sprintf(day,"0%d",var2->tm_mday);
        else                    sprintf(day,"%d",var2->tm_mday);
        if((var2->tm_mon+1)<10) sprintf(month,"0%d",var2->tm_mon+1);
        else                    sprintf(month,"%d",var2->tm_mon+1);
        sprintf(year,"%d",var2->tm_year+1900);
        if(var2->tm_hour<10)    sprintf(hr,"0%d",var2->tm_hour);
        else                    sprintf(hr,"%d",var2->tm_hour);
        if(var2->tm_min<10)     sprintf(min,"0%d",var2->tm_min);
        else                    sprintf(min,"%d",var2->tm_min);
        if(var2->tm_sec<10)     sprintf(sec,"0%d",var2->tm_sec);
        else                    sprintf(sec,"%d",var2->tm_sec);

        sprintf(buffer, "%s-%s-%s %s:%s:%s - ClownIDS - %s",day,month,year,hr,min,sec,buff);

        fprintf(logfile,"%s\n",buffer);
        fflush(logfile);
}

void makesnapfiles()
{
	char path[128];
	char buf[256];
	int i,j,x;

	memset(path,0,128);
	for(i=0;i<128;i++) memset(snap[i],0,128);
	strncpy(path,DB_PREFIX,strlen(DB_PREFIX));

	if(path[strlen(path)-1] == '/') path[strlen(path)-1] = 0;

	for(i=0;i<n;i++)
	{
		memset(buf,0,256);x=0;
		for(j=0;j<strlen(data[i]);j++) if(data[i][j]!='/') buf[x++] = data[i][j];
		sprintf(snap[i],"%s/%s.snap",path,buf);
	}
}

void snapshot(int sel)
{
	int i;
	char cmd[2048];
	char buf[128];

	if(sel == 1)
	{
		sprintf(logbuf,"Taking md5 snapshot ...");
		log(logbuf);
	}

	for(i=0;i<n;i++)
	{
		memset(buf,0,128);
		strncpy(buf,snap[i],strlen(snap[i]));
		if(sel==2) strncat(buf,".chk",4);
		chdir(data[i]);
		sprintf(cmd, "md5 `ls` >%s",buf);
		system(cmd);
	}
}

unsigned long filesize(FILE *stream)
{
	unsigned long curpos, length;

	curpos = ftell(stream);
	fseek(stream, 0L, SEEK_END);
	length = ftell(stream);
	fseek(stream, curpos, SEEK_SET);
	return length;
}

int checksnapshot()
{
	int i;
	unsigned long fs = 0;
	FILE *fp;
	char tmp[] = "/tmp/.idsXXXXXX";
	char path[128];
	char cmd[2048];

	mkstemp(tmp);

	snapshot(2);

	for(i=0;i<n;i++)
	{
		sprintf(cmd,"diff %s %s.chk >>%s",snap[i],snap[i],tmp);
		system(cmd);
	}

	fp=fopen(tmp,"r");
	fs = filesize(fp);
	fclose(fp);

	if(fs>0)
	{
		sprintf(logbuf,"*** INTRUSION DETECTED *** logs mailed to %s",ADMIN_MAIL);log(logbuf);
		/* mail the differences to the admin */
		sprintf(cmd, "echo \"`cat %s`\" |mail -s \"%s\" %s",tmp,SUBJECT,ADMIN_MAIL);
		system(cmd);
		remove(tmp);

	        /* clean up .chk files */
	        memset(path,0,128);
		strncpy(path,DB_PREFIX,strlen(DB_PREFIX));
		if(path[strlen(path)-1] == '/') path[strlen(path)-1] = 0;
		sprintf(cmd,"rm -f %s/*.chk",path);system(cmd);

		return 1;
	}

	remove(tmp);

	/* clean up .chk files */
	memset(path,0,128);
	strncpy(path,DB_PREFIX,strlen(DB_PREFIX));
	if(path[strlen(path)-1] == '/') path[strlen(path)-1] = 0;
	sprintf(cmd,"rm -f %s/*.chk",path);system(cmd);

	return 0;
}

int unlocked()
{
	FILE *fp;

	if((fp=fopen(KEY_FILE,"r")) !=NULL)
	{
		fclose(fp);
		return 0;
	}

	return 1;
}

int main()
{
	int i;

	banner();

        if((logfile=fopen(IDS_LOG,"a"))==NULL)
        {
                printf("  Error: couldn't write to logfile %s.\n\n",IDS_LOG);
                return 1;
        }

	if(getuid()!=0)
	{
		printf("  Error: Only root can run ClownIDS.\n\n");
		sprintf(logbuf,"Attempt to run ClownIDS as uid %d [halted]\n",getuid());
		log(logbuf);
		return 1;
	}

	if(!readconfig())
	{
		printf("  Error: ClownIDS aborting.\n\n");
		return 1;
	}

	if(VERBOSE)
	{
		showdirs();
		printf("   Init: Logging to: %s\n",IDS_LOG);
	}

        printf("   Init: Taking md5 snapshot ...\n");
	makesnapfiles();
	snapshot(1);

	printf("   Init: ClownIDS fork()'ing.\n\n");

	if(fork()==0)
	{
		while(1)
		{
			for(i=0;i<60;i++)
			{
				if(unlocked()==0)
				{
					sprintf(logbuf,"Keyfile detected - ClownIDS paused");log(logbuf);

					while(unlocked()==0) sleep(5);

					sprintf(logbuf,"Keyfile removed - ClownIDS unpaused");log(logbuf);

					snapshot(1);
					i=60;
				}

				sleep(INTERVAL);
			}

			snapshot(2);
			if(checksnapshot()==1)
			{
				/* EMERGENCY */

				system(EMER_CMD);
				return 0;
			}
		}
	}

	return 0;
}
